﻿using System.Collections.Generic;
using HEAL.Attic;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;

namespace HeuristicLab.Problems.DataAnalysis.Symbolic.Regression {
  [Item("Constraint ConstOpt Evaluator", "")]
  [StorableType("4170FD8B-DDD9-43B7-8BF8-F9C7290D4D1C")]
  public class SymbolicRegressionSingleObjectiveConstraintConstOptEvaluator : SymbolicRegressionSingleObjectiveEvaluator {
    [StorableConstructor]
    protected SymbolicRegressionSingleObjectiveConstraintConstOptEvaluator(StorableConstructorFlag _) : base(_) { }

    protected SymbolicRegressionSingleObjectiveConstraintConstOptEvaluator(
      SymbolicRegressionSingleObjectiveConstraintConstOptEvaluator original, Cloner cloner)
      : base(original, cloner) { }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new SymbolicRegressionSingleObjectiveConstraintConstOptEvaluator(this, cloner);
    }

    public SymbolicRegressionSingleObjectiveConstraintConstOptEvaluator() { }

    public override bool Maximization => true;

    public override IOperation InstrumentedApply() {
      var solution = SymbolicExpressionTreeParameter.ActualValue;
      var rows     = GenerateRowsToEvaluate();

      var quality = Calculate(SymbolicDataAnalysisTreeInterpreterParameter.ActualValue, solution,
                              EstimationLimitsParameter.ActualValue.Lower, EstimationLimitsParameter.ActualValue.Upper,
                              ProblemDataParameter.ActualValue, rows, ApplyLinearScalingParameter.ActualValue.Value);
      QualityParameter.ActualValue = new DoubleValue(quality);
      return base.InstrumentedApply();
    }

    public static double Calculate(ISymbolicDataAnalysisExpressionTreeInterpreter interpreter,
                                   ISymbolicExpressionTree tree, double lowerEstimationLimit,
                                   double upperEstimationLimit, IRegressionProblemData problemData,
                                   IEnumerable<int> rows, bool applyLinearScaling) {
      var estimatedValues = interpreter.GetSymbolicExpressionTreeValues(tree, problemData.Dataset, rows);
      var targetValues    = problemData.Dataset.GetDoubleValues(problemData.TargetVariable, rows);

      var constraints    = problemData.IntervalConstraints.EnabledConstraints;
      var variableRanges = problemData.VariableRanges.GetDictionary();

      if (!SymbolicRegressionConstraintAnalyzer.ConstraintsSatisfied(constraints, variableRanges, tree)) return 0;

      var r2 = SymbolicRegressionConstantOptimizationEvaluator.OptimizeConstants(interpreter, tree, problemData,
                                                                                 problemData.TrainingIndices,
                                                                                 applyLinearScaling, 10);
      return r2;
    }

    public override double Evaluate(IExecutionContext context, ISymbolicExpressionTree tree,
                                    IRegressionProblemData problemData, IEnumerable<int> rows) {
      SymbolicDataAnalysisTreeInterpreterParameter.ExecutionContext = context;
      EstimationLimitsParameter.ExecutionContext                    = context;
      ApplyLinearScalingParameter.ExecutionContext                  = context;

      var r2 = Calculate(SymbolicDataAnalysisTreeInterpreterParameter.ActualValue, tree,
                         EstimationLimitsParameter.ActualValue.Lower, EstimationLimitsParameter.ActualValue.Upper,
                         problemData, rows,
                         ApplyLinearScalingParameter.ActualValue.Value);

      SymbolicDataAnalysisTreeInterpreterParameter.ExecutionContext = null;
      EstimationLimitsParameter.ExecutionContext                    = null;
      ApplyLinearScalingParameter.ExecutionContext                  = null;

      return r2;
    }
  }
}